# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
from . import pipeline
from ..utilities import helpers
from .. import datatable


class Batch(object):

    """Run a pipeline batch process.

    Args:
    * `source`: path to the source file
    * `source_type`: 'csv' or 'datapackage'
    * `data_key`: The key for the data (only used when source_type is 'csv')
    * `schema_key`: The key for the schema (only used when source_type is 'csv').
    * `pipeline_options`: Options to pass to each pipeline instance.
    * `post_task`: Task handler to run after all pipelines have run.
    * `pipeline_post_task`: Task handler passed to each pipeline instance.

    """

    def __init__(self, source, source_type='csv', data_key='data',
                 schema_key='schema', pipeline_options=None,
                 post_task=None, pipeline_post_task=None):

        self.source = source
        self.source_type = source_type
        self.data_key = data_key
        self.schema_key = schema_key
        self.dataset = self.get_dataset()
        self.pipeline_options = pipeline_options or {}
        self.current_pipeline = None
        self.reports = []

        helpers.validate_handler(post_task, 1)
        helpers.validate_handler(pipeline_post_task, 1)

        self.post_task = post_task
        self.pipeline_post_task = pipeline_post_task

    def get_dataset(self):
        """Get the dataset for this batch process."""

        return getattr(self, 'get_dataset_{0}'.format(self.source_type))()

    def get_dataset_csv(self):
        """Get the dataset from a CSV file for this batch process."""

        dataset = []
        resources = datatable.DataTable(self.source)

        data_index = resources.headers.index(self.data_key)
        schema_index = None
        if self.schema_key:
            schema_index = resources.headers.index(self.schema_key)

        for entry in resources.values:

            rv = {'data': entry[data_index], 'schema': None}

            if schema_index is not None:
                rv['schema'] = entry[schema_index]

            dataset.append(rv)

        return dataset

    def get_dataset_datapackage(self):
        """Get the dataset from a Data Package for this batch process."""

        _name = 'datapackage.json'
        descriptor = os.path.join(self.source, _name)
        dataset = []
        # TODO: We want to use https://github.com/tryggvib/datapackage here
        # but, in order to do so, we need these issues resolved:
        # https://github.com/tryggvib/datapackage/issues/35
        # https://github.com/tryggvib/datapackage/issues/33
        pkg = helpers.load_json_source(descriptor)
        for entry in pkg['resources']:

            if entry.get('url'):
                data = entry['url']
            elif entry.get('path'):
                if pkg.get('base'):
                    data = '{0}{1}'.format(pkg['base'], entry['path'])
                else:
                    data = os.path.join(self.source, entry['path'])
            else:
                data = entry.get('data')

            dataset.append({
                'data': data,
                'schema': entry.get('schema')
            })

        return dataset

    def pipeline_factory(self, data, schema):
        """Construct a pipeline."""

        options = self.pipeline_options or {}
        if options.get('options') is None:
            options['options'] = {}

        if options['options'].get('schema') is None:
            options['options']['schema'] = {}

        options['options']['schema']['schema'] = schema

        return pipeline.Pipeline(data, post_task=self.pipeline_post_task,
                                 **options)

    def run(self):
        """Run the batch."""

        # TODO: parallelize

        for data in self.dataset:
            self.current_pipeline = self.pipeline_factory(data['data'],
                                                          data['schema'])
            result, report = self.current_pipeline.run()
            self.reports.append(report)

        if self.post_task:
            self.post_task(self)

        return not any([report.count for report in self.reports])



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import jtskit
from . import base
from .. import exceptions


RESULTS = {
    'schema_001': {
        'id': 'schema_001',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema. Data has {0}, but they should be {1}.',
        'help': '',
        'help_edit': ''
    },
    'schema_002': {
        'id': 'schema_002',
        'name': 'Incorrect Dimensions',
        'msg': 'The row dimensions do not match the header dimensions.',
        'help': '',
        'help_edit': ''
    },
    'schema_003': {
        'id': 'schema_003',
        'name': 'Incorrect Type',
        'msg': 'The value is not a valid {0}.',
        'help': '',
        'help_edit': ''
    },
    'schema_004': {
        'id': 'schema_004',
        'name': 'Required Field',
        'msg': 'Column {0} is a required field, but no value can be found in row {1}.',
        'help': '',
        'help_edit': ''
    },
    'schema_005': {
        'id': 'schema_005',
        'name': 'Non-Required Field (Empty/Null)',
        'msg': 'Column {0} is a non-required field, and has a null value in row {1}.',
        'help': '',
        'help_edit': ''
    },
    'schema_006': {
        'id': 'schema_006',
        'name': 'Unique Field',
        'msg': 'Column {0} is a unique field, yet the value {1} already exists.',
        'help': '',
        'help_edit': ''
    }
}


class SchemaProcessor(base.Processor):

    """Process data against a JSON Table Schema."""

    name = 'schema'
    RESULT_TYPES = RESULTS

    def __init__(self, fail_fast=False, transform=False, report_limit=1000,
                 row_limit=30000, schema=None, ignore_field_order=True,
                 report_stream=None, report=None,
                 result_level='error', infer_schema=False,
                 case_insensitive_headers=False, **kwargs):

        super(SchemaProcessor, self).__init__(
            fail_fast=fail_fast, transform=transform,
            report_limit=report_limit, row_limit=row_limit,
            report_stream=report_stream, report=report, result_level=result_level)

        self.infer_schema = infer_schema
        self.case_insensitive_headers = case_insensitive_headers
        self.ignore_field_order = ignore_field_order
        if not schema:
            self.schema = None
        else:
            self.schema = self.schema_model(schema)

        self._uniques = {}

    def schema_model(self, schema):
        try:
            model = jtskit.models.SchemaModel(schema, self.case_insensitive_headers)
        except (jtskit.exceptions.InvalidJSONError, jtskit.exceptions.InvalidSchemaError) as e:
            raise exceptions.ProcessorBuildError(e.msg)

        return model

    def pre_run(self, data_table):

        if (self.schema is None) and self.infer_schema:
            sample_values = data_table.get_sample(300)
            self.schema = self.schema_model(jtskit.infer(data_table.headers, sample_values))

        return True, data_table

    def run_header(self, headers, header_index=0):

        valid = True

        if self.case_insensitive_headers:
            headers = [name.lower() for name in headers]

        if self.schema:
            if self.ignore_field_order:
                if not (set(headers).issuperset(set(self.schema.required_headers))):

                    valid = False
                    _type = RESULTS['schema_001']
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_HEADER,
                        self.RESULT_LEVEL_ERROR,
                        _type['msg'].format(headers, self.schema.headers),
                        _type['id'],
                        _type['name'],
                        headers,
                        header_index,
                        self.RESULT_HEADER_ROW_NAME
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers

            else:
                header_length = len(headers)
                if not (headers == self.schema.required_headers[:header_length]):

                    valid = False
                    _type = RESULTS['schema_001']
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_HEADER,
                        self.RESULT_LEVEL_ERROR,
                        _type['msg'].format(headers, self.schema.headers),
                        _type['id'],
                        _type['name'],
                        headers,
                        header_index,
                        self.RESULT_HEADER_ROW_NAME,
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers

        return valid, headers

    def run_row(self, headers, index, row):

        valid = True
        row_name = self.get_row_id(headers, row)

        if self.schema:
            if not (len(headers) == len(row)):

                valid = False
                _type = RESULTS['schema_002']
                entry = self.make_entry(
                    self.name,
                    self.RESULT_CATEGORY_ROW,
                    self.RESULT_LEVEL_ERROR,
                    _type['msg'],
                    _type['id'],
                    _type['name'],
                    row,
                    index,
                    row_name,
                )

                self.report.write(entry)
                if self.fail_fast:
                    return valid, headers, index, row

            else:
                for column_name, column_value in zip(headers, row):

                    # handle case where column_name not even in schema
                    if not self.schema.has_field(column_name):
                        pass

                    # we know the field is in the schema
                    else:
                        # check type and format
                        if self.schema.cast(column_name, column_value) is False:

                            valid = False
                            _type = RESULTS['schema_003']
                            entry = self.make_entry(
                                self.name,
                                self.RESULT_CATEGORY_ROW,
                                self.RESULT_LEVEL_ERROR,
                                _type['msg'].format(self.schema.get_type(column_name).name.title()),
                                _type['id'],
                                _type['name'],
                                row,
                                index,
                                row_name,
                                headers.index(column_name),
                                column_name
                            )

                            self.report.write(entry)
                            if self.fail_fast:
                                return valid, headers, index, row

                        # CONSTRAINTS
                        constraints = self.schema.get_constraints(column_name)

                        if constraints['required'] is True and \
                           (column_value in self.schema.NULL_VALUES):

                            valid = False
                            _type = RESULTS['schema_004']
                            entry = self.make_entry(
                                self.name,
                                self.RESULT_CATEGORY_ROW,
                                self.RESULT_LEVEL_ERROR,
                                _type['msg'].format(column_name, index),
                                _type['id'],
                                _type['name'],
                                row,
                                index,
                                row_name,
                                headers.index(column_name),
                                column_name
                            )

                            self.report.write(entry)
                            if self.fail_fast:
                                return valid, headers, index, row

                        if constraints['required'] is False and \
                           (column_value in self.schema.NULL_VALUES) and \
                           self.result_level == self.RESULT_LEVEL_INFO:
                            # add info result
                            _type = RESULTS['schema_005']
                            entry = self.make_entry(
                                self.name,
                                self.RESULT_CATEGORY_ROW,
                                self.RESULT_LEVEL_INFO,
                                _type['msg'].format(column_name, index),
                                _type['id'],
                                _type['name'],
                                row,
                                index,
                                row_name,
                                headers.index(column_name),
                                column_name
                            )

                            self.report.write(entry)
                            if self.fail_fast:
                                return valid, headers, index, row

                        if constraints.get('unique') is True:

                            if not self._uniques.get(column_name):
                                self._uniques[column_name] = set([column_value])

                            elif column_value in self._uniques[column_name]:
                                _type = RESULTS['schema_006']
                                entry = self.make_entry(
                                    self.name,
                                    self.RESULT_CATEGORY_ROW,
                                    self.RESULT_LEVEL_ERROR,
                                    _type['msg'].format(column_name, column_value),
                                    _type['id'],
                                    _type['name'],
                                    row,
                                    index,
                                    row_name,
                                    headers.index(column_name),
                                    column_name
                                )

                                self.report.write(entry)
                                if self.fail_fast:
                                    return valid, headers, index, row

                            else:
                                self._uniques[column_name].add(column_value)

                        # TODO: check constraints.min* and constraints.max*

        return valid, headers, index, row



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import json
import click
from goodtables import pipeline as _pipeline
from goodtables import processors


@click.group()
def cli():
    """The entry point into the CLI."""


@cli.command()
@click.argument('data')
@click.option('--schema')
@click.option('--format', default='csv', type=click.Choice(['csv', 'excel']))
@click.option('--dry_run', is_flag=True)
@click.option('--fail_fast', is_flag=True)
@click.option('--row_limit', default=20000, type=int)
@click.option('--report_limit', default=1000, type=int)
def pipeline(data, schema, format, dry_run, fail_fast, row_limit, report_limit):

    """Run a Good Tables pipeline."""

    options = {}
    if schema:
        options['schema'] = {'schema': schema}

    processor = _pipeline.Pipeline(data, format=format, fail_fast=fail_fast,
                                   dry_run=dry_run, row_limit=row_limit,
                                   report_limit=report_limit)

    valid, report = processor.run()

    click.echo(json.dumps(report.generate(), ensure_ascii=True))


@cli.command()
@click.argument('data')
@click.option('--format', default='csv', type=click.Choice(['csv', 'excel']))
@click.option('--fail_fast', is_flag=True)
@click.option('--row_limit', default=20000, type=int)
@click.option('--report_limit', default=1000, type=int)
@click.option('--output', default='txt', type=click.Choice(['txt', 'json']))
def structure(data, format, fail_fast, row_limit, report_limit, output):

    """Run a Good Tables StructureProcessor."""

    processor = processors.StructureProcessor(format=format, fail_fast=fail_fast,
                                              row_limit=row_limit,
                                              report_limit=report_limit)

    valid, report, data = processor.run(data)

    valid_msg = 'Well done! The data is valid :)\n'.upper()
    invalid_msg = 'Oops.The data is invalid :(\n'.upper()

    if output == 'json':
        exclude = None
    else:
        exclude = ['result_context', 'processor', 'row_name', 'result_category',
                   'column_index', 'column_name', 'result_level']

    if valid:
        click.echo(click.style(valid_msg, fg='green'))
    else:
        click.echo(click.style(invalid_msg, fg='red'))

    click.echo(report.generate(output, exclude=exclude))

@cli.command()
@click.argument('data')
@click.option('--schema')
@click.option('--format', default='csv', type=click.Choice(['csv', 'excel']))
@click.option('--fail_fast', is_flag=True)
@click.option('--row_limit', default=20000, type=int)
@click.option('--report_limit', default=1000, type=int)
@click.option('--output', default='txt', type=click.Choice(['txt', 'json']))
def schema(data, schema, format, fail_fast, row_limit, report_limit, output):

    """Run a Good Tables SchemaProcessor."""

    processor = processors.SchemaProcessor(schema=schema, format=format,
                                           fail_fast=fail_fast, row_limit=row_limit,
                                           report_limit=report_limit)

    valid, report, data = processor.run(data)

    if output == 'json':
        exclude = None
    else:
        exclude = ['result_context', 'processor', 'row_name', 'result_category',
                   'column_index', 'column_name', 'result_level']

    valid_msg = 'Well done! The data is valid :)\n'.upper()
    invalid_msg = 'Oops.The data is invalid :(\n'.upper()

    if valid:
        click.echo(click.style(valid_msg, fg='green'))
    else:
        click.echo(click.style(invalid_msg, fg='red'))

    click.echo(report.generate(output, exclude=exclude))


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .base import Processor
from .structure import StructureProcessor
from .schema import SchemaProcessor


__all__ = ['Processor', 'StructureProcessor', 'SchemaProcessor']



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import io
import json
import jsonschema
from . import helpers


def validate(source):
    """Validate a CSV Dialect source file."""

    schemafile = os.path.abspath(os.path.join(
        os.path.dirname(os.path.dirname(__file__)), 'schemas',
        'csv-dialect-description-format.json'))

    with io.open(schemafile) as stream:
        schema = json.load(stream)

    try:
        source = helpers.load_json_source(source)
        jsonschema.validate(source, schema)
        return True

    except (jsonschema.ValidationError, ValueError, TypeError):
        return False



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import pipeline
from . import processors
from . import utilities
from . import exceptions
from . import compat


__all__ = ['pipeline', 'processors', 'utilities', 'exceptions', 'compat']



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import io
import importlib
import shutil
import tempfile
import json
import csv
import decimal
import tellme
from ..utilities import data_package, csv_dialect, helpers
from .. import datatable
from .. import exceptions
from .. import compat


class Pipeline(object):

    """Validate a (tabular) data source through a validation pipeline.

    Args:
    * processors: A list of processor names to process `data_source`
        * Each name can be a 'shortname' for the default processors
            * e.g., ['structure', 'schema']
        * Each name can be a string path to a processor
            * e.g., ['custompackage.CustomValidator', 'schema']
            * Custom processors must implement the Validator API
    * data: A buffer, filepath, string or URL to the table data
    * format: The format of `data_source`. 'csv' or 'json'
    * dialect: A buffer, filepath, string or URL to a CSV dialect spec
    * options: a dict configuration object for the pipeline
        * Each processor has its options nested under its 'shortname'
        * Custom processors have options nested under cls.__name__.lower()
        * e.g.:
            {'structure': {#options}, 'customprocessor': {#options}}
    * workspace: path to directory for files. e.g.: '/my/path'
    * dry_run: No files are persisted after the run has been completed

    Returns:
        A tuple of `valid, report`, where `valid` is a boolean expressing
        validity according to the whole pipeline, and report is a dict
        of the entire pipeline report where each top-level key matches
        a processor in the pipeline.

    """

    ROW_LIMIT_MAX = 30000
    REPORT_LIMIT_MAX = 1000
    SOURCE_FILENAME = 'source.csv'
    TRANSFORM_FILENAME = 'transform.csv'
    DIALECT_FILENAME = 'dialect.json'
    DATA_FORMATS = ('csv', 'excel', 'json')

    def __init__(self, data, processors=None, dialect=None, format='csv',
                 encoding=None, options=None, workspace=None, dry_run=True,
                 transform=True, fail_fast=False, row_limit=20000,
                 report_limit=1000, report_stream=None, header_index=0,
                 break_on_invalid_processor=True, post_task=None,
                 report_post_task=None):

        if data is None:
            _msg = '`data` must be a filepath, url or stream.'
            raise exceptions.PipelineBuildError(_msg)

        self.openfiles = []
        self.processors = processors or helpers.DEFAULT_PIPELINE
        self.dialect = self.get_dialect(dialect)
        self.format = format
        self.encoding = encoding
        self.options = options or {}
        self.dry_run = dry_run
        self.workspace = self.get_workspace(workspace)
        self.transform = transform
        self.fail_fast = fail_fast
        self.row_limit = self.get_row_limit(row_limit)
        self.report_limit = self.get_report_limit(report_limit)
        self.report_stream = report_stream
        self.header_index = header_index
        self.break_on_invalid_processor = break_on_invalid_processor

        helpers.validate_handler(report_post_task)
        helpers.validate_handler(post_task)

        self.report_post_task = report_post_task or helpers.pipeline_stats
        self.post_task = post_task

        if self.report_stream:
            report_stream_tests = [isinstance(self.report_stream, io.TextIOBase),
                                   self.report_stream.writable(),
                                   self.report_stream.seekable()]

            if not all(report_stream_tests):
                _msg = '`report_stream` must be a seekable and writable text stream.'
                raise exceptions.PipelineBuildError(_msg)

            report_backend = 'client'
        else:
            report_backend = 'yaml'

        report_options = {
            'schema': helpers.report_schema,
            'backend': report_backend,
            'client_stream': self.report_stream,
            'limit': self.report_limit,
            'post_task': self.report_post_task
        }

        self.report = tellme.Report('Pipeline', **report_options)

        self.pipeline = self.get_pipeline()

        try:
            self.data = datatable.DataTable(data, format=self.format,
                                            encoding=encoding,
                                            header_index=self.header_index)
        except datatable.DataTable.RAISES as e:
            raise exceptions.PipelineBuildError(e.msg)

        self.openfiles.extend(self.data.openfiles)

        if not self.dry_run:
            self.init_workspace()

    def init_workspace(self):
        """Initalize the workspace for this run."""

        self.create_file(self.data.stream, self.SOURCE_FILENAME, self.data.headers)
        self.create_file('', self.TRANSFORM_FILENAME)
        self.create_file(json.dumps(self.dialect), self.DIALECT_FILENAME)

    def get_workspace(self, filepath):
        """Return a workspace for this run."""
        # TODO: use pudo/barn

        if not self.dry_run:
            return filepath or tempfile.mkdtemp()

        return None

    def get_pipeline(self):
        """Get the pipeline for this instance."""

        pipeline = []
        for name in self.processors:
            _class = self.resolve_processor(name)
            options = self.options.get(_class.name, {})
            # override options with those we accept directly from the pipeline
            options['row_limit'] = self.row_limit
            options['report_limit'] = self.report_limit
            options['report_stream'] = self.report_stream
            options['transform'] = self.transform
            options['fail_fast'] = self.fail_fast
            options['header_index'] = self.header_index
            options['report'] = self.report

            try:
                instance = _class(**options)
                pipeline.append(instance)
            except _class.RAISES as e:
                raise exceptions.PipelineBuildError(e.msg)

        return pipeline

    def get_dialect(self, dialect_source):
        """Get a CSV dialect instance for this data."""

        if dialect_source:
            _valid, rv = csv_dialect.validate(dialect_source)
            if not _valid:
                raise exceptions.InvalidSpec
            return rv
        else:
            return None

    def get_row_limit(self, passed_limit):
        """Return the row limit, locked to an upper limit."""

        if passed_limit > self.ROW_LIMIT_MAX:
            return self.ROW_LIMIT_MAX
        else:
            return passed_limit

    def get_report_limit(self, passed_limit):
        """Return the report_limit, locked to an upper limit."""

        if passed_limit > self.REPORT_LIMIT_MAX:
            return self.REPORT_LIMIT_MAX
        else:
            return passed_limit

    def create_file(self, data, name, headers=None):
        """Create a file in the pipeline workspace."""

        filepath = os.path.join(self.workspace, name)
        with io.open(filepath, mode='w+t', encoding='utf-8') as destfile:
            if headers:
                destfile.write(','.join(headers))

            if isinstance(data, compat.str):
                destfile.write(data)
            elif isinstance(data, compat.bytes):
                # TODO: We should not ever deal with bytes here: see why we are (or if we still are)
                destfile.write(data.decode('utf-8'))
            else:
                destfile.write(data.read())
                data.seek(0)

    def resolve_processor(self, processor_name):
        """Return a processor class."""

        if processor_name in helpers.builtin_processors():
            processor_class = helpers.builtin_processors()[processor_name]
        else:
            # resolve a custom processor
            _module, _class = processor_name.rsplit('.', 1)
            try:
                processor_class = getattr(importlib.import_module(_module),
                                          _class)
            except ImportError as e:
                # TODO: something better here
                raise e

        return processor_class

    def register_processor(self, processor_name, options=None, position=None):
        """Register a processor on the pipeline."""

        processor_class = self.resolve_processor(processor_name)
        options = options or {}
        processor = processor_class(**options)

        if position is None:
            self.pipeline.append(processor)
        else:
            self.pipeline.insert(position, processor)

    def run(self):
        """Run the pipeline."""

        def _run_valid(process_valid, run_valid):
            """Set/maintain the valid state of the run."""
            if not process_valid and run_valid:
                return False
            return run_valid

        valid = True

        for processor in self.pipeline:

            _valid, _, self.data = processor.run(self.data, is_table=True)
            valid = _run_valid(_valid, valid)

            # if a validator returns invalid, we stop the pipeline,
            # unless break_on_invalid_processor is False
            if not valid and self.break_on_invalid_processor:
                break

            if not self.dry_run and self.transform:
                # TODO: do what we'll do with transform data, workspace, etc.
                pass
            else:
                self.data.replay()

        self.set_report_meta()

        if self.post_task:
            # TODO: handle what happens in here
            self.post_task(self)

        return valid, self.report

    def rm_workspace(self):
        """Remove this run's workspace from disk."""

        return shutil.rmtree(self.workspace)

    def set_report_meta(self):
        """Set information and statistics for this run on report['meta']."""
        self.report.meta['row_count'] = self.pipeline[0].row_count or 1
        self.report.meta['header_index'] = self.header_index
        self.report.meta['headers'] = self.data.headers



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import io
import json
import inspect
import requests
from .. import compat
from .. import exceptions


def builtin_processors():
    """Return dict of public builtin processors. Avoids circular import."""
    from .. import processors
    return {
        processors.StructureProcessor.name: processors.StructureProcessor,
        processors.SchemaProcessor.name: processors.SchemaProcessor
    }


DEFAULT_PIPELINE = ('structure',)
REMOTE_SCHEMES = ('http', 'https', 'ftp', 'ftps')


# a schema for the reporter.Report() instances used by validators
report_schema = {
    'processor': {'type': compat.str},
    'result_category': {'type': compat.str},
    'result_level': {'type': compat.str},
    'result_message': {'type': compat.str},
    'result_id': {'type': compat.str},
    'result_name': {'type': compat.str},
    'result_context': {'type': list},
    'row_index': {'type': (int, type(None))},
    'row_name': {'type': compat.str},
    'column_index': {'type': (int, type(None))},
    'column_name': {'type': compat.str}
}


def get_report_result_types():
    """Return a list of all builtin result types."""

    result_types = []

    for processor in builtin_processors().values():
        result_types.extend([r[1] for r in
                             tuple(processor.RESULT_TYPES.items())])

    return result_types


def pipeline_stats(report):
    """Generates high-level pipeline statistics from a report."""

    def _bad_type_percent(column_name, results, row_count):
        """Return a percentage for amount of data with bad type in a column."""

        match_count = len([r for r in results if
                           r['result_id'] == 'schema_003' and
                           r['column_name'] == column_name])

        return int((match_count/row_count) * 100)

    bad_row_count = len(set([r['row_index'] for r in report['results'] if
                             r['result_category'] == 'row' and
                             r['result_level'] == 'error' and
                             r['row_index'] is not None]))

    bad_column_count = len(set([r['column_index'] for r in report['results'] if
                                r['result_category'] == 'row' and
                                r['result_level'] == 'error' and
                                r['column_index'] is not None]))

    columns = [{'name': header, 'index': index,
                'bad_type_percent': _bad_type_percent(header, report['results'],
                                                      report['meta']['row_count'])}
               for index, header in enumerate(report['meta']['headers'])]

    report['meta'].update({
        'bad_row_count': bad_row_count,
        'bad_column_count': bad_column_count,
        'columns': columns
    })

    return report


def load_json_source(source):

    """Load a JSON source, from string, URL or buffer,  into a Python type."""

    if source is None:
        return None

    elif isinstance(source, (dict, list)):
        # the source has already been loaded
        return source

    elif compat.parse.urlparse(source).scheme in REMOTE_SCHEMES:
        return json.loads(requests.get(source).text)

    elif isinstance(source, compat.str) and not os.path.exists(source):
        return json.loads(source)

    else:
        with io.open(source, encoding='utf-8') as stream:
            source = json.load(io.open(source, encoding='utf-8'))
        return source


def validate_handler(handler, argcount=1):

    """Validate a handler.

    Args:
        * `handler`: the supposed handler, which is a function
        * `argcount`: the count of args the handler should take
        (not including self if handler is a method)
    """

    if handler is None:
        return True
    else:
        if not callable(handler):
            raise exceptions.InvalidHandlerError

        if inspect.ismethod(handler):
            # extra arg for self
            argcount += 1

        spec = inspect.getargspec(handler)
        if not len(spec[0]) == argcount:
            raise exceptions.InvalidHandlerError
